#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []  # type: ignore

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]

setup(
    author="Alex de Ruijter",
    author_email="alex@slicerisk.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    description="A package to deal with risk matrix data.",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="riskmatrix",
    name="riskmatrix",
    packages=find_packages(include=["riskmatrix"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/slicerisk/riskmatrix",
    version="0.0.9",
    zip_safe=False,
)
